"""
    verktyg_server.tests.test_sockets
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright:
        (c) 2015 by Ben Mather.
    :license:
        BSD, see LICENSE for more details.
"""
import unittest


class SocketsTestCase(unittest.TestCase):
    pass
